import abc
from past.builtins import cmp


class Server(object):

    def __init__(self, address, port, weight=1, priority=1, ttl=300):
        self.address = address
        self.port = port
        self.weight = weight
        self.priority = priority
        self.ttl = ttl
        self._is_alive = False

    def __str__(self):
        return "%s(%s:%s, ttl:%s, weight:%s, priority:%s, alive:%s)" % (
            self.__class__.__name__,
            self.address,
            self.port,
            self.ttl,
            self.weight,
            self.priority,
            self.is_alive
        )

    def __hash__(self):
        return hash((self.address, self.port))

    def __eq__(self, other):
        return (
            self.address == other.address and
            self.port == other.port
        )

    def __lt__(self, other):
        return self.priority < other.priority

    def __cmp__(self, other):
        return cmp(self.priority, other.priority)

    @property
    def is_alive(self):
        return self._is_alive


class ServerList(object):

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_servers(self):
        return []


class ServerStats(object):

    @property
    def active_requests(self):
        """
        The current number of active requests
        """
        return 0

    @property
    def utilization(self):
        """
        The current utilization as a percentage between 0 and 100.
        """
        return 0

    @property
    def failure_count(self):
        return 0

    @property
    def average_response_time(self):
        return 0

    def add_response_time(self, time):
        pass

    def increment_failures(self):
        pass

    def increment_active_requests(self):
        pass

    def decrement_active_requests(self):
        pass

    def is_tripped(self, current_time):
        """
        Whether or not the circuit breaker tripped due to too many failures.
        """
        return False



import abc
import logging
import threading
import socket
import requests
from timeit import default_timer as timer
from multiprocessing.pool import ThreadPool, Pool
from balast.discovery import Server, ServerList
from balast.exception import BalastException


# current thread name (so we know who 'main' is)
_MAIN_THREAD_NAME = threading.currentThread().name

# the thread name format for our ping threads
_THREAD_NAME_FORMAT = 'balast-ping-'


def _ping_in_background(ping, server):

    # rename the thread according to our convention
    t = threading.currentThread()
    if not t.name == _MAIN_THREAD_NAME and not t.name.startswith('balast'):
        t.setName(_THREAD_NAME_FORMAT + t.name)

    # ping the server
    server._is_alive = ping.is_alive(server)

    return server


class Ping(object):

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.max_ping_time = None
        self._logger = logging.getLogger(self.__module__)

    def __getstate__(self):

        # logger can't be pickeled
        d = self.__dict__.copy()
        del d['_logger']

        return d

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._logger = logging.getLogger(self.__module__)

    @abc.abstractmethod
    def is_alive(self, server):
        return False


class DummyPing(Ping):

    def is_alive(self, server):

        assert isinstance(server, Server)

        self._logger.debug("Ping succeeded for server: %s", server)

        return True


class SocketPing(Ping):

    def is_alive(self, server):
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(self.max_ping_time)
            s.connect((server.address, server.port))
            s.close()

            self._logger.debug("Ping succeeded for server: %s", server)

            return True
        except:
            self._logger.warn("Ping failed for server: %s", server)
            return False


class UrlPing(Ping):

    _HTTP = 'http'
    _HTTPS = 'https'
    _URL_FORMAT = '{}://{}:{}'

    def __init__(self, is_secure=False):
        super(UrlPing, self).__init__()
        self.is_secure = is_secure

    def is_alive(self, server):
        try:
            scheme = self._HTTPS \
                if self.is_secure \
                else self._HTTP

            url = self._URL_FORMAT.format(
                scheme,
                server.address,
                server.port
            )

            result = requests.get(url)

            self._logger.debug("Ping succeeded for server: %s", server)

            return result.ok
        except:
            self._logger.warn("Ping failed for server: %s", server)
            return False


class PingStrategy(object):

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self._logger = logging.getLogger(self.__module__)

    def __getstate__(self):

        # logger can't be pickeled
        d = self.__dict__.copy()
        del d['_logger']

        return d

    def __setstate__(self, state):
        self.__dict__.update(state)

    @abc.abstractmethod
    def ping(self, ping, servers):
        return []


class SerialPingStrategy(PingStrategy):

    def ping(self, ping, servers):

        assert isinstance(ping, Ping)
        assert isinstance(servers, ServerList)

        start_time = timer()

        # returns a list of booleans
        results = []
        for s in servers.get_servers():
            is_alive = ping.is_alive(s)
            s._is_alive = is_alive
            results.append(s)

        end_time = timer() - start_time

        self._logger.debug("Pinged %s servers in %s seconds", len(results), end_time)

        return results


class AsyncPoolPingStrategy(PingStrategy):

    def __init__(self, pool_type):
        super(AsyncPoolPingStrategy, self).__init__()
        self._pool_type = pool_type

    def __getstate__(self):
        s = super(AsyncPoolPingStrategy, self).__getstate__()
        return s

    def __setstate__(self, state):
        pass

    def ping(self, ping, servers):

        assert isinstance(ping, Ping)
        assert isinstance(servers, ServerList)

        start_time = timer()

        futures = []
        results = []
        server_list = list(servers.get_servers())
        if not any(server_list):
            self._logger.debug("Ping failed - no servers to ping!")
            return results

        pool = self._pool_type(processes=len(server_list))

        # create our threads in the thread pool
        for s in server_list:
            future = pool.apply_async(_ping_in_background, (ping, s))
            futures.append(future)

        # now, join the threads and grab the results
        for f in futures:
            server = f.get()
            results.append(server)

        pool.close()

        end_time = timer() - start_time

        self._logger.debug("Pinged %s servers in %s seconds", len(results), end_time)

        return results


class ThreadPoolPingStrategy(AsyncPoolPingStrategy):

    def __init__(self):
        super(ThreadPoolPingStrategy, self).__init__(ThreadPool)


class MultiprocessingPoolPingStrategy(AsyncPoolPingStrategy):

    def __init__(self):
        super(MultiprocessingPoolPingStrategy, self).__init__(Pool)


class GeventPingStrategy(PingStrategy):

    def ping(self, ping, servers):

        assert isinstance(ping, Ping)
        assert isinstance(servers, ServerList)

        start_time = timer()

        try:
            import gevent
        except ImportError:
            raise BalastException(
                "Please install optional gevent dependencies "
                "in order to use this feature: \n\n"
                "$ pip install balast[gevent] or \n"
                "$ pip install balast[all]"
            )

        results = []
        greenlets = []

        # create our greenlets
        for s in servers.get_servers():
            g = gevent.spawn(_ping_in_background, ping, s)
            greenlets.append(g)

        # wait for all greenlets to finish
        gevent.joinall(greenlets)

        # now, grab the results
        for g in greenlets:
            server = g.get()
            results.append(server)

        end_time = timer() - start_time

        self._logger.debug("Pinged %s servers in %s seconds", len(results), end_time)

        return results



from past.builtins import basestring


class BalastException(Exception):

    _DEFAULT_MSG = 'An unexpected error occurred'

    def __init__(self, msg=None, cause=None):

        assert msg is None or isinstance(msg, basestring)
        assert cause is None or isinstance(cause, BaseException)

        self.cause = cause

        message = msg
        if message is None and cause is not None:
            message = cause.message
        elif message is None and cause is None:
            message = self._DEFAULT_MSG

        super(BalastException, self).__init__(message)


class NoReachableServers(BalastException):

    _DEFAULT_MSG = 'No reachable servers found!'


class BalastConfigurationException(BalastException):
    pass



import logging
from past.builtins import basestring
from balast.discovery import Server, ServerList
from balast.exception import BalastException


class StaticServerList(ServerList):

    def __init__(self, servers):
        self._servers = set()
        self._logger = logging.getLogger(self.__module__)

        for server in servers:

            if isinstance(server, Server):
                self._servers.add(server)
            elif isinstance(server, basestring):
                parts = server.split(":")

                if len(parts) > 1:
                    self.add_server(parts[0], int(parts[1]))
                else:
                    self.add_server(parts[0])
            else:
                raise BalastException('Server was in unexpected format: "%s"' % server)

    def add_server(self, address, port=80, weight=1, priority=1):
        s = Server(address, port, weight, priority)
        self._servers.add(s)

    def get_servers(self):
        self._logger.debug("Resolved %s servers", len(self._servers))
        return set(self._servers)



import abc
import threading
from queue import Queue
from balast.discovery import Server
from balast.exception import BalastException, NoReachableServers


class Rule(object):

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self._load_balancer = None

    @property
    def load_balancer(self):
        return self._load_balancer

    @load_balancer.setter
    def load_balancer(self, value):
        self._load_balancer = value

    @abc.abstractmethod
    def choose(self):
        return Server(None, None)


class RoundRobinRule(Rule):

    def __init__(self):
        super(RoundRobinRule, self).__init__()
        self._lock = threading.Lock()
        self._queue = Queue()

    def choose(self):

        with self._lock:

            if self._load_balancer is None:
                raise BalastException("Load balancer not set!")

            # push all available servers back onto
            # the queue if the queue is empty
            if self._queue.empty():
                servers = self._load_balancer.reachable_servers
                if len(servers) == 0:
                    raise NoReachableServers()
                sorted_by_priority = sorted(servers)
                for server in sorted_by_priority:
                    self._queue.put(server)

            # now just pop one off until
            # the queue is empty
            return self._queue.get(block=False)


class PriorityWeightedRule(Rule):

    def __init__(self):
        super(PriorityWeightedRule, self).__init__()
        self._load_balancer_stats = None

    @property
    def load_balancer(self):
        return self._load_balancer

    @load_balancer.setter
    def load_balancer(self, value):
        self._load_balancer = value
        self._load_balancer_stats = value.stats

    def choose(self):

        servers = self._load_balancer.reachable_servers

        server_count = len(servers)

        if len(servers) == 0:
            raise Exception("No reachable servers could be found")

        if server_count == 1:
            return servers.pop()

        def s(o1, o2):
            return -1

        sorted_by_priority = sorted(servers, s)

        # for server in sorted_by_priority:
        # stats = self._load_balancer_stats.get_server_stats(server)
        # assert isinstance(stats, ServerStats)
        # TODO: check weight/priority

        # sort the servers by priority
        # from the top-priority servers, get the server with utilization lowest compared to weight
        # if all utilizations are lower than respective weights, just pick the first server

        return sorted_by_priority[0]



import logging
import threading
import time
from balast.discovery import ServerList
from balast.rule import Rule, RoundRobinRule
from balast.ping import (
    Ping,
    SocketPing,
    PingStrategy,
    SerialPingStrategy
)


class LoadBalancer(object):

    DEFAULT_PING_INTERVAL = 30
    MAX_PING_TIME = 3

    def __init__(self, server_list, rule=None, ping_strategy=None, ping=None, ping_on_start=True):

        assert isinstance(server_list, ServerList)
        assert rule is None or isinstance(rule, Rule)
        assert ping_strategy is None or isinstance(ping_strategy, PingStrategy)
        assert ping is None or isinstance(ping, Ping)

        # some locks for thread-safety
        self._lock = threading.Lock()
        self._server_lock = threading.Lock()

        self._rule = rule \
            if rule is not None \
            else RoundRobinRule()

        self._ping_strategy = ping_strategy \
            if ping_strategy is not None \
            else SerialPingStrategy()

        self._ping = ping \
            if ping is not None \
            else SocketPing()

        self.max_ping_time = self.MAX_PING_TIME
        self._ping_interval = self.DEFAULT_PING_INTERVAL
        self._server_list = server_list
        self._servers = set()
        self._stats = LoadBalancerStats()
        self._rule.load_balancer = self
        self._logger = logging.getLogger(self.__module__)

        # start our background worker
        # to periodically ping our servers
        self._ping_timer_running = False
        self._ping_timer = None
        if ping_on_start:
            self._start_ping_timer()

    @property
    def ping_interval(self):
        return self._ping_interval

    @ping_interval.setter
    def ping_interval(self, value):
        self._ping_interval = value

        if self._ping_timer_running:
            self._stop_ping_timer()
            self._start_ping_timer()

    @property
    def max_ping_time(self):
        if self._ping is None:
            return 0
        return self._ping.max_ping_time

    @max_ping_time.setter
    def max_ping_time(self, value):
        if self._ping is not None:
            self._ping.max_ping_time = value

    @property
    def stats(self):
        return self._stats

    @property
    def servers(self):
        with self._server_lock:
            return set(self._servers)

    @property
    def reachable_servers(self):
        with self._server_lock:
            servers = set()
            for s in self._servers:
                if s.is_alive:
                    servers.add(s)

        return servers

    def choose_server(self):

        # choose a server, will
        # throw if there are none
        server = self._rule.choose()

        return server

    def mark_server_down(self, server):
        self._logger.debug("Marking server down: %s", server)
        server._is_alive = False

    def ping(self, server=None):
        if server is None:
            self._ping_all_servers()
        else:
            is_alive = self._ping.is_alive(server)
            server._is_alive = is_alive

    def ping_async(self, server=None):
        if server is None:
            # self._ping_all_servers()
            t = threading.Thread(name='balast-worker', target=self._ping_all_servers)
            t.daemon = True
            t.start()
        else:
            is_alive = self._ping.is_alive(server)
            server._is_alive = is_alive

    def _ping_all_servers(self):
        with self._server_lock:
            results = self._ping_strategy.ping(
                self._ping,
                self._server_list
            )
            self._servers = set(results)

    def _start_ping_timer(self):

        with self._lock:
            if self._ping_timer_running:
                self._logger.debug("Background pinger already running")
                return

            self._ping_timer_running = True
            self._ping_timer = threading.Thread(name='balast-worker', target=self._ping_loop)
            self._ping_timer.daemon = True
            self._ping_timer.start()

    def _stop_ping_timer(self):
        with self._lock:
            self._ping_timer_running = False
            self._ping_timer = None

    def _ping_loop(self):
        while self._ping_timer_running:
            try:
                self._ping_all_servers()
            except BaseException as e:
                self._logger.error("There was an error pinging servers: %s", e)

            time.sleep(self._ping_interval)


class LoadBalancerStats(object):

    def get_server_stats(self, server):
        pass



from past.builtins import basestring, unicode
try:
    from urllib.parse import urlparse, urlunparse, urljoin, parse_qs, urlencode
except ImportError:
    from urlparse import urlparse, urlunparse, urljoin, parse_qs
    from urllib import urlencode


class UrlBuilder(object):

    DEFAULT_SCHEME = 'http'
    DEFAULT_PORT = 80

    def __init__(self):
        self._scheme = ''
        self._hostname = ''
        self._port = ''
        self._path = ''
        self._query = dict()
        self._username = ''
        self._password = ''
        self._fragment = ''

    @staticmethod
    def from_url(url):
        parts = urlparse(url)
        return UrlBuilder.from_parts(
            parts.scheme,
            parts.username,
            parts.password,
            parts.hostname,
            parts.port,
            parts.path,
            parts.query,
            parts.fragment
        )

    @staticmethod
    def from_parts(
            scheme=DEFAULT_SCHEME,
            username='',
            password='',
            hostname='',
            port=None,
            path='',
            query='',
            fragment=''
    ):

        builder = UrlBuilder().\
            scheme(scheme).\
            username(username).\
            password(password).\
            hostname(hostname).\
            port(port).\
            path(path).\
            fragment(fragment)

        if query is not None:
            if isinstance(query, basestring):
                builder._query = parse_qs(query)
            elif isinstance(query, dict):
                # TODO: validate structure?
                builder._query = query
            else:
                raise Exception('Query format unexpected!')

        return builder

    def __repr__(self):
        return "%s(%s)" % (
            self.__class__.__name__,
            self.build()
        )

    def __str__(self):
        return self.build()

    def __unicode__(self):
        return self.build()

    def scheme(self, value):

        assert value is None or isinstance(value, basestring)

        self._scheme = value
        return self

    def http(self):
        self._scheme = 'http'
        return self

    def https(self):
        self._scheme = 'https'
        return self

    def hostname(self, value):

        assert value is None or isinstance(value, basestring)

        self._hostname = value if value is not None else ''
        return self

    def port(self, value):

        assert value is None or isinstance(value, int)

        self._port = value
        return self

    def path(self, value):

        assert value is None or isinstance(value, basestring)

        self._path = value
        return self

    def username(self, value):

        assert value is None or isinstance(value, basestring)

        self._username = value
        return self

    def password(self, value):

        assert value is None or isinstance(value, basestring)

        self._password = value
        return self

    def fragment(self, value):

        assert value is None or isinstance(value, basestring)

        self._fragment = value
        return self

    def add_query_param(self, key, value):

        assert key is None or isinstance(key, basestring)

        if key not in self._query:
            self._query[key] = []

        self._query[key].append(unicode(value))

        return self

    def remove_query_param(self, key, value=None):

        assert key is None or isinstance(key, basestring)

        # nothing to do if the key isn't in there
        if key not in self._query:
            return self

        # if value is None, remove all
        # params for the specified key
        if value is None:
            del self._query[key]
            return self

        # otherwise, just remove the specified
        # value from the query param list for
        # the specified key
        l = self._query[key]
        l.remove(unicode(value))

        # if there are no more values,
        # remove the key from the dictionary
        if len(l) == 0:
            del self._query[key]

        return self

    def append_path(self, path):

        assert path is None or isinstance(path, basestring)

        base = self._path

        if not base.endswith('/'):
            base += '/'

        self._path = urljoin(base, path)

        return self

    def build(self):
        # create the array the unparse method expects
        # and populate with our values
        parts = [''] * 6
        parts[0] = self._scheme
        parts[1] = self._build_host()
        parts[2] = self._path
        parts[4] = urlencode(self._query, 1)
        parts[5] = self._fragment

        # finally, create the url from the parts
        return urlunparse(parts)

    def _build_host(self):

        if self._username is not None and self._username != '':
            host = '{}:{}@{}'.format(
                self._username,
                self._password,
                self._hostname
            )
        else:
            host = self._hostname

        if self._port is not None and self._port != self.DEFAULT_PORT:
            host += ':{}'.format(self._port)

        return host



import logging
import requests
from past.builtins import basestring, unicode
from requests.exceptions import RequestException
from balast.util import UrlBuilder
from balast.core import LoadBalancer
from balast.exception import BalastConfigurationException
from balast.discovery import ServerList
from balast.discovery.static import StaticServerList


class Service(object):

    DEFAULT_REQUEST_TIMEOUT = 10

    def __init__(self, *args, **kwargs):
        self._load_balancer = kwargs.get('load_balancer')
        self._use_https = kwargs.get('use_https', False)
        self._request_timeout = kwargs.get('request_timeout', self.DEFAULT_REQUEST_TIMEOUT)
        self._logger = logging.getLogger(self.__module__)

        # if our load balancer wasn't configured via kwargs
        # and there are no positional args, we have a problem
        if self._load_balancer is None and len(args) == 0:
            raise BalastConfigurationException(
                "Expected either a collection of server "
                "addresses OR a balast.LoadBalancer "
                "instance as an init parameter."
            )

        # either the load balancer is specified or the collection, not both
        if self._load_balancer is not None and len(args) > 0:
            raise BalastConfigurationException(
                "Keyword arg 'load_balancer' specified in addition to positional arg. "
                "Please either remove the positional arg, or specify the load balancer "
                "as the positional arg."
            )

        # load balancer or collection of servers
        # can be defined via 1st positional arg
        # (but only if load balancer not set via kwargs)
        a = args[0]
        if isinstance(a, LoadBalancer):
            self._load_balancer = a
        elif isinstance(a, ServerList):
            servers = a
            self._load_balancer = LoadBalancer(servers)
        elif hasattr(a, '__iter__') and not isinstance(a, basestring):
            servers = StaticServerList(a)
            self._load_balancer = LoadBalancer(servers)
        else:
            raise BalastConfigurationException(
                "An invalid configuration parameter was provided: %s" %
                a
            )

    def request(self, method, url, **kwargs):

        # choose a server from the pool
        server = self._load_balancer.choose_server()
        absolute_url = self._get_absolute_url(server, url, self._use_https)

        self._logger.debug("Request: %s %s", method.upper(), absolute_url)

        try:
            response = requests.request(method, absolute_url, timeout=self._request_timeout, **kwargs)

            # 5xx errors should mark the server down
            # everything else is good to go
            if response.status_code < 500:
                return response

        except RequestException as e:
            self._logger.error("Request to server failed for url: '%s': %s", absolute_url, e)

        # mark this server down and try the
        # request again with a new server
        self._load_balancer.mark_server_down(server)

        return self.request(method, url, **kwargs)

    def options(self, url, **kwargs):

        # choose a server from the pool
        server = self._load_balancer.choose_server()
        absolute_url = self._get_absolute_url(server, url, self._use_https)

        self._logger.debug("Request: OPTIONS %s", absolute_url)

        try:
            response = requests.options(absolute_url, timeout=self._request_timeout, **kwargs)

            # 5xx errors should mark the server down
            # everything else is good to go
            if response.status_code < 500:
                return response

        except RequestException as e:
            self._logger.error("Request to server failed for url: '%s': %s", absolute_url, e)

        # mark this server down and try the
        # request again with a new server
        self._load_balancer.mark_server_down(server)

        return self.options(url, **kwargs)

    def head(self, url, **kwargs):

        # choose a server from the pool
        server = self._load_balancer.choose_server()
        absolute_url = self._get_absolute_url(server, url, self._use_https)

        self._logger.debug("Request: HEAD %s", absolute_url)

        try:
            response = requests.head(absolute_url, timeout=self._request_timeout, **kwargs)

            # 5xx errors should mark the server down
            # everything else is good to go
            if response.status_code < 500:
                return response

        except RequestException as e:
            self._logger.error("Request to server failed for url: '%s': %s", absolute_url, e)

        # mark this server down and try the
        # request again with a new server
        self._load_balancer.mark_server_down(server)

        return self.head(url, **kwargs)

    def get(self, url, params=None, **kwargs):

        # choose a server from the pool
        server = self._load_balancer.choose_server()
        absolute_url = self._get_absolute_url(server, url, self._use_https)

        self._logger.debug("Request: GET %s", absolute_url)

        try:
            response = requests.get(absolute_url, params, timeout=self._request_timeout, **kwargs)

            # 5xx errors should mark the server down
            # everything else is good to go
            if response.status_code < 500:
                return response

        except RequestException as e:
            self._logger.error("Request to server failed for url: '%s': %s", absolute_url, e)

        # mark this server down and try the
        # request again with a new server
        self._load_balancer.mark_server_down(server)

        return self.get(url, params, **kwargs)

    def post(self, url, data=None, json=None, **kwargs):

        # choose a server from the pool
        server = self._load_balancer.choose_server()
        absolute_url = self._get_absolute_url(server, url, self._use_https)

        self._logger.debug("Request: POST %s", absolute_url)

        try:
            response = requests.post(absolute_url, data, json, timeout=self._request_timeout, **kwargs)

            # 5xx errors should mark the server down
            # everything else is good to go
            if response.status_code < 500:
                return response

        except RequestException as e:
            self._logger.error("Request to server failed for url: '%s': %s", absolute_url, e)

        # mark this server down and try the
        # request again with a new server
        self._load_balancer.mark_server_down(server)

        return self.post(url, data, json, **kwargs)

    def put(self, url, data=None, **kwargs):

        # choose a server from the pool
        server = self._load_balancer.choose_server()
        absolute_url = self._get_absolute_url(server, url, self._use_https)

        self._logger.debug("Request: PUT %s", absolute_url)

        try:
            response = requests.put(absolute_url, data, timeout=self._request_timeout, **kwargs)

            # 5xx errors should mark the server down
            # everything else is good to go
            if response.status_code < 500:
                return response

        except RequestException as e:
            self._logger.error("Request to server failed for url: '%s': %s", absolute_url, e)

        # mark this server down and try the
        # request again with a new server
        self._load_balancer.mark_server_down(server)

        return self.put(url, data, **kwargs)

    def patch(self, url, data=None, **kwargs):

        # choose a server from the pool
        server = self._load_balancer.choose_server()
        absolute_url = self._get_absolute_url(server, url, self._use_https)

        self._logger.debug("Request: PATCH %s", absolute_url)

        try:
            response = requests.patch(absolute_url, data, timeout=self._request_timeout, **kwargs)

            # 5xx errors should mark the server down
            # everything else is good to go
            if response.status_code < 500:
                return response

        except RequestException as e:
            self._logger.error("Request to server failed for url: '%s': %s", absolute_url, e)

        # mark this server down and try the
        # request again with a new server
        self._load_balancer.mark_server_down(server)

        return self.patch(url, data, **kwargs)

    def delete(self, url, **kwargs):

        # choose a server from the pool
        server = self._load_balancer.choose_server()
        absolute_url = self._get_absolute_url(server, url, self._use_https)

        self._logger.debug("Request: DELETE %s", absolute_url)

        try:
            response = requests.delete(absolute_url, timeout=self._request_timeout, **kwargs)

            # 5xx errors should mark the server down
            # everything else is good to go
            if response.status_code < 500:
                return response

        except RequestException as e:
            self._logger.error("Request to server failed for url: '%s': %s", absolute_url, e)

        # mark this server down and try the
        # request again with a new server
        self._load_balancer.mark_server_down(server)

        return self.delete(url, **kwargs)

    @staticmethod
    def _get_absolute_url(server, relative_url, use_https):

        url = UrlBuilder.from_parts(
            hostname=server.address,
            port=server.port
        )
        url.append_path(relative_url)

        if use_https:
            url.https()

        return unicode(url)



from balast.core import LoadBalancer
from balast.ping import Ping, PingStrategy
from balast.rule import Rule
from balast.discovery import Server, ServerList
from balast.service import Service

import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())

__all__ = [
    'LoadBalancer',
    'Ping',
    'PingStrategy',
    'Rule',
    'Server',
    'ServerList',
    'Service'
]



import unittest
from balast import LoadBalancer
from balast.discovery.static import StaticServerList
from balast.rule import RoundRobinRule
from balast.ping import Ping, DummyPing
from balast.exception import BalastException


class _MockPing(Ping):

    def __init__(self, is_alive):
        super(_MockPing, self).__init__()
        self._is_alive = is_alive

    def is_alive(self, server):
        return self._is_alive


class RoundRobinRuleTest(unittest.TestCase):

    def test_choose_without_setting_balancer(self):
        rule = RoundRobinRule()
        self.assertRaises(BalastException, rule.choose)

    def test_equal_choice(self):

        servers = StaticServerList(['127.0.0.1', '127.0.0.2', '127.0.0.3'])
        load_balancer = LoadBalancer(servers, ping=DummyPing(), ping_on_start=False)

        # ping our servers once to set them available
        load_balancer.ping()

        rule = RoundRobinRule()
        rule.load_balancer = load_balancer

        stats = dict()

        expected_iterations = 1000

        # each should be chosen an equal number of times
        # loop a bunch to get some stats
        for i in range(expected_iterations):
            server1 = rule.choose()
            if server1 in stats:
                stats[server1] += 1
            else:
                stats[server1] = 1

            server2 = rule.choose()
            if server2 in stats:
                stats[server2] += 1
            else:
                stats[server2] = 1

            server3 = rule.choose()
            if server3 in stats:
                stats[server3] += 1
            else:
                stats[server3] = 1

        # all 3 should have been chosen the same number of times
        self.assertEqual(3, len(stats))
        for server in stats:
            self.assertEqual(expected_iterations, stats[server])

    def test_no_servers_reachable(self):

        servers = StaticServerList(['127.0.0.1', '127.0.0.2', '127.0.0.3'])
        load_balancer = LoadBalancer(servers, ping=_MockPing(False), ping_on_start=False)

        rule = RoundRobinRule()
        rule.load_balancer = load_balancer

        self.assertRaises(BalastException, rule.choose)



import unittest
from past.builtins import unicode
from balast.util import UrlBuilder
try:
    from urllib.parse import urlparse, parse_qs
except ImportError:
    from urlparse import urlparse, parse_qs


class UrlBuilderTest(unittest.TestCase):

    def test_builder_url_should_be_same(self):

        expected_url = 'http://test.com/path/to/resource'
        url = UrlBuilder.from_url(expected_url)

        self.assertEqual(expected_url, unicode(url))

        expected_url = 'http://my-user:my-pass@test.com:8080/some/path?test=hi&test=hi2&other=bye#my-frag/path/to/frag'
        url = UrlBuilder.from_url(expected_url)

        actual_url = urlparse(unicode(url))
        actual_query = parse_qs(actual_url.query)

        self.assertEqual(actual_url.scheme, 'http')
        self.assertEqual(actual_url.username, 'my-user')
        self.assertEqual(actual_url.password, 'my-pass')
        self.assertEqual(actual_url.hostname, 'test.com')
        self.assertEqual(actual_url.path, '/some/path')
        self.assertEqual(actual_url.fragment, 'my-frag/path/to/frag')
        self.assertIn('hi', actual_query['test'])
        self.assertIn('hi2', actual_query['test'])
        self.assertIn('bye', actual_query['other'])



#!/usr/bin/env python

import re
import os
from setuptools import setup, find_packages, Command


def get_version():

    # create a dummy file if it doesn't exist
    if not os.path.exists('version.py'):
        with open('version.py', 'w') as f:
            f.write("__version__ = '0.0.0'\n")

    # now open and parse the version number
    with open('version.py') as f:
        contents = f.read()
        version = re.search(r"__version__\s*=\s*'(.*)'", contents, re.M).group(1)
        return version


class GenerateVersionCommand(Command):
    """
    Create a `version.py` file containing the
    specified version information.
    """
    description = 'generate version number file'

    user_options = [
        ('sha', None, 'include commit SHA hash'),
    ]

    def initialize_options(self):
        self.sha = False

    def finalize_options(self):
        self.sha = True if int(self.sha) == 1 else False

    def run(self):

        major, minor, revision, offset, sha = self._get_version_from_git()

        print("Major: %s" % major)
        print("Minor: %s" % minor)
        print("Revision: %s" % revision)
        print("Offset: %s" % offset)
        print("Sha: %s" % sha)

        with open('version.py', 'w') as f:
            if self.sha:
                f.write("__version__ = '%s.%s.%s-%s'\n" % (major, minor, offset, sha))
            else:
                f.write("__version__ = '%s.%s.%s'\n" % (major, minor, offset))

    @staticmethod
    def _get_version_from_git():
        import re
        import git

        # parse our version number
        # from a combination of the
        # git revision count as well
        # as git tagged  version
        repo = git.Repo()
        head = repo.head.commit
        revision = head.count()
        d = repo.git.describe(long=True, tags=True)
        m = re.search('^v([0-9]*)\.([0-9]*)\.[0-9]*-([0-9]*)-(.*)$', d)

        assert m is not None

        g = m.groups()

        return g[0], g[1], revision, g[2], g[3]


with open('README.rst', 'r') as f:
    readme = f.read()


setup(name='balast',
      version=get_version(),
      description='Balast client-side load-balancing framework',
      long_description=readme,
      author='Justin Smith',
      author_email='smith.justin.c@gmail.com',
      maintainer='Justin Smith',
      maintainer_email='smith.justin.c@gmail.com',
      license='MIT',
      url='https://github.com/radishllc/balast',
      packages=find_packages(exclude=['test', 'docs']),
      package_data={
          'balast': ['../version.py', '../LICENSE'],
      },
      include_package_data=True,
      setup_requires=['pytest-runner'],
      tests_require=['pytest', 'mock'],
      install_requires=['requests', 'future'],
      extras_require={
          'dns': ['dnspython'],
          'gevent': ['gevent'],
          'all': ['dnspython', 'gevent']
      },
      cmdclass={
          'version': GenerateVersionCommand
      },
      classifiers=[
          # How mature is this project? Common values are
          #   3 - Alpha
          #   4 - Beta
          #   5 - Production/Stable
          # "Development Status :: 1 - Planning",
          # "Development Status :: 2 - Pre-Alpha",
          "Development Status :: 3 - Alpha",
          # "Development Status :: 4 - Beta",

          # Indicate who your project is intended for
          'Intended Audience :: Developers',
          'Topic :: Software Development :: Libraries :: Python Modules',

          # Pick your license as you wish (should match "license" above)
          'License :: OSI Approved :: MIT License',

          # Specify the Python versions you support here. In particular, ensure
          # that you indicate whether you support Python 2, Python 3 or both.
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          # 'Programming Language :: Python :: 3.2',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy'
      ],
      keywords='alpha development load-balancing',
      )



import abc
import logging
import socket
from past.builtins import unicode
from balast.exception import BalastException, BalastConfigurationException
from balast.discovery import ServerList, Server

try:
    from dns import resolver, rdatatype, exception
    from dns.rdtypes.IN.A import A
    from dns.rdtypes.ANY.CNAME import CNAME
except ImportError:
    raise BalastException(
        "Please install optional DNS dependencies "
        "in order to use this feature: \n\n"
        "$ pip install balast[dns] or \n"
        "$ pip install balast[all]"
    )


class DnsRecordList(ServerList):

    __metaclass__ = abc.ABCMeta

    _RESOLVER_CACHE = resolver.LRUCache()

    def __init__(self, dns_qname, dns_host=None, dns_port=None):

        self._dns_qname = dns_qname
        self._logger = logging.getLogger(self.__module__)

        # create a DNS resolver that caches results
        self._dns_resolver = resolver.Resolver()
        self._dns_resolver.cache = DnsRecordList._RESOLVER_CACHE

        if dns_port is not None:
            self._dns_resolver.port = dns_port

        if dns_host is not None:
            try:
                self._dns_resolver.nameservers = [
                    socket.gethostbyname(dns_host)
                ]
            except Exception as e:
                raise BalastConfigurationException(e.message, e)

    @abc.abstractmethod
    def get_servers(self):
        pass


class DnsARecordList(DnsRecordList):

    def __init__(self, dns_qname, dns_host=None, dns_port=None, server_port=80):
        super(DnsARecordList, self).__init__(
            dns_qname,
            dns_host,
            dns_port
        )
        self.server_port = server_port

    def get_servers(self):

        try:
            # query SRV records for our service name
            answer = self._dns_resolver.query(self._dns_qname, rdatatype.A)

            # iterate the results, generate server objects
            for i, srv in enumerate(answer):
                ttl = answer.response.answer[0].ttl
                s = Server(
                    srv.address,
                    self.server_port,
                    ttl=ttl
                )

                self._logger.debug("Created server from DNS A record: %s", s)

                yield s

        except (exception.DNSException, BalastException):
            return


class DnsServiceRecordList(DnsRecordList):

    def __init__(self, dns_qname, dns_host=None, dns_port=None):
        super(DnsServiceRecordList, self).__init__(
            dns_qname,
            dns_host,
            dns_port
        )

    def get_servers(self):

        try:
            # query SRV records for our service name
            answer = self._dns_resolver.query(self._dns_qname, rdatatype.SRV)

            # iterate the results, generate server objects
            for i, srv in enumerate(answer):
                rdata = answer.response.additional[0].items[i]
                if isinstance(rdata, A):
                    address = rdata.address
                elif isinstance(rdata, CNAME):
                    address = unicode(rdata.target).rstrip('.')
                else:
                    raise BalastException('Unexpected DNS record: %s' % rdata)

                ttl = answer.response.additional[0].ttl
                s = Server(
                    address,
                    srv.port,
                    srv.weight,
                    srv.priority,
                    ttl
                )

                self._logger.debug("Created server from DNS SRV record: %s", s)

                yield s

        except (exception.DNSException, BalastException):
            return



import unittest
import mock
from requests import models, exceptions
from balast import ping, Service, LoadBalancer
from balast.util import UrlBuilder
from balast.discovery import Server
from balast.discovery.static import StaticServerList
from balast.exception import (
    BalastException,
    NoReachableServers,
    BalastConfigurationException
)


_EXPECTED_SERVERS = ['127.0.0.1', '127.0.0.2']


class _MockResponse(models.Response):

    def __init__(self, status_code):
        super(_MockResponse, self).__init__()
        self.status_code = status_code


class ServiceTest(unittest.TestCase):

    def setUp(self):
        self._service = self._create_service()
        self._load_balancer = self._service._load_balancer
        self._servers = self._load_balancer.servers

    def test_init_with_server_collection(self):

        servers = [
            Server('127.0.0.1', 80),
            Server('127.0.0.2', 81)
        ]
        # create a service with a collection
        # of server addresses
        service = Service(servers)

        # should have auto-created a load balancer
        load_balancer = service._load_balancer
        self.assertIsNotNone(load_balancer)

        # should have auto-created a static server list
        server_list = load_balancer._server_list
        self.assertIsInstance(server_list, StaticServerList)

        for server in server_list.get_servers():
            self.assertIn(server.address, _EXPECTED_SERVERS)

    def test_init_with_server_string_collection(self):

        # create a service with a collection
        # of server addresses
        service = Service(_EXPECTED_SERVERS)

        # should have auto-created a load balancer
        load_balancer = service._load_balancer
        self.assertIsNotNone(load_balancer)

        # should have auto-created a static server list
        server_list = load_balancer._server_list
        self.assertIsInstance(server_list, StaticServerList)

        for server in server_list.get_servers():
            self.assertIn(server.address, _EXPECTED_SERVERS)

    def test_init_with_bad_args(self):

        # no args
        self.assertRaises(BalastConfigurationException, lambda: Service())

        # positional arg not LoadBalancer or collection
        self.assertRaises(BalastConfigurationException, lambda: Service('bad-arg!'))

        # collection items not in proper format (either Server or strings)
        self.assertRaises(BalastException, lambda: Service([1, 2, 3]))

        # positional arg AND LoadBalancer
        def init_service():
            Service('bad-arg!', load_balancer=self._load_balancer)
        self.assertRaises(BalastConfigurationException, init_service)

    @mock.patch('balast.service.requests.request', return_value=_MockResponse(200))
    def test_request_ok(self, mock_request):

        # make our request
        result = self._service.request('GET', '/relative/path')

        # should have tried twice - once for each server
        self.assertEqual(1, mock_request.call_count)
        self.assertEqual(result, mock_request.return_value)

        # verify the url we create
        call_arg = mock_request.call_args_list[0]
        url = UrlBuilder.from_url(call_arg[0][1])
        self.assertIn(url._hostname, _EXPECTED_SERVERS)
        self.assertEqual(url._scheme, 'https')
        self.assertIsNone(url._port)
        self.assertEqual(url._path, '/relative/path')

    @mock.patch('balast.service.requests.head', return_value=_MockResponse(200))
    def test_head_ok(self, mock_request):
        self.assert_request_ok(mock_request, self._service.head)

    @mock.patch('balast.service.requests.options', return_value=_MockResponse(200))
    def test_options_ok(self, mock_request):
        self.assert_request_ok(mock_request, self._service.options)

    @mock.patch('balast.service.requests.get', return_value=_MockResponse(200))
    def test_get_ok(self, mock_request):
        self.assert_request_ok(mock_request, self._service.get)

    @mock.patch('balast.service.requests.post', return_value=_MockResponse(200))
    def test_post_ok(self, mock_request):
        self.assert_request_ok(mock_request, self._service.post)

    @mock.patch('balast.service.requests.put', return_value=_MockResponse(200))
    def test_put_ok(self, mock_request):
        self.assert_request_ok(mock_request, self._service.put)

    @mock.patch('balast.service.requests.patch', return_value=_MockResponse(200))
    def test_patch_ok(self, mock_request):
        self.assert_request_ok(mock_request, self._service.patch)

    @mock.patch('balast.service.requests.delete', return_value=_MockResponse(200))
    def test_delete_ok(self, mock_request):
        self.assert_request_ok(mock_request, self._service.delete)

    @mock.patch('balast.service.requests.request', return_value=_MockResponse(500))
    def test_all_request_servers_error(self, mock_request):

        mock_request.side_effect = exceptions.RequestException('mock exception')

        # will try each server, and mark each down on a 500 response
        # when no servers are left, a NoReachableServersException is thrown
        self.assertRaises(NoReachableServers, self._service.request, 'GET', '/relative/path')

        # should have tried twice - once for each server
        self.assertEqual(2, mock_request.call_count)

        # add our hostnames to a set (to ensure the
        # same server isn't being used repeatedly)
        # all other info should be the same for each server request
        hostnames = set()
        for i, call_arg in enumerate(mock_request.call_args_list):
            url = UrlBuilder.from_url(call_arg[0][1])
            hostnames.add(url._hostname)
            self.assertEqual(url._scheme, 'https')
            self.assertIsNone(url._port)
            self.assertEqual(url._path, '/relative/path')

        # less than 2 means a server has been reused
        self.assertEqual(2, len(hostnames))
        for hostname in hostnames:
            self.assertIn(hostname, _EXPECTED_SERVERS)

    @mock.patch('balast.service.requests.options', return_value=_MockResponse(500))
    def test_all_options_servers_error(self, mock_request):
        mock_request.side_effect = exceptions.RequestException('mock exception')
        self.assert_all_servers_fail(mock_request, self._service.options)

    @mock.patch('balast.service.requests.head', return_value=_MockResponse(500))
    def test_all_head_servers_error(self, mock_request):
        mock_request.side_effect = exceptions.RequestException('mock exception')
        self.assert_all_servers_fail(mock_request, self._service.head)

    @mock.patch('balast.service.requests.get', return_value=_MockResponse(500))
    def test_all_get_servers_error(self, mock_request):
        mock_request.side_effect = exceptions.RequestException('mock exception')
        self.assert_all_servers_fail(mock_request, self._service.get)

    @mock.patch('balast.service.requests.post', return_value=_MockResponse(500))
    def test_all_post_servers_error(self, mock_request):
        mock_request.side_effect = exceptions.RequestException('mock exception')
        self.assert_all_servers_fail(mock_request, self._service.post)

    @mock.patch('balast.service.requests.put', return_value=_MockResponse(500))
    def test_all_put_servers_error(self, mock_request):
        mock_request.side_effect = exceptions.RequestException('mock exception')
        self.assert_all_servers_fail(mock_request, self._service.put)

    @mock.patch('balast.service.requests.patch', return_value=_MockResponse(500))
    def test_all_patch_servers_error(self, mock_request):
        mock_request.side_effect = exceptions.RequestException('mock exception')
        self.assert_all_servers_fail(mock_request, self._service.patch)

    @mock.patch('balast.service.requests.delete', return_value=_MockResponse(500))
    def test_all_delete_servers_error(self, mock_request):
        mock_request.side_effect = exceptions.RequestException('mock exception')
        self.assert_all_servers_fail(mock_request, self._service.delete)

    @mock.patch('balast.service.requests.request', return_value=_MockResponse(500))
    def test_all_request_servers_500(self, mock_request):

        # will try each server, and mark each down on a 500 response
        # when no servers are left, a NoReachableServersException is thrown
        self.assertRaises(NoReachableServers, self._service.request, 'GET', '/relative/path')

        # should have tried twice - once for each server
        self.assertEqual(2, mock_request.call_count)

        # add our hostnames to a set (to ensure the
        # same server isn't being used repeatedly)
        # all other info should be the same for each server request
        hostnames = set()
        for i, call_arg in enumerate(mock_request.call_args_list):
            url = UrlBuilder.from_url(call_arg[0][1])
            hostnames.add(url._hostname)
            self.assertEqual(url._scheme, 'https')
            self.assertIsNone(url._port)
            self.assertEqual(url._path, '/relative/path')

        # less than 2 means a server has been reused
        self.assertEqual(2, len(hostnames))
        for hostname in hostnames:
            self.assertIn(hostname, _EXPECTED_SERVERS)

    @mock.patch('balast.service.requests.options', return_value=_MockResponse(500))
    def test_all_options_servers_500(self, mock_request):
        self.assert_all_servers_fail(mock_request, self._service.options)

    @mock.patch('balast.service.requests.head', return_value=_MockResponse(500))
    def test_all_head_servers_500(self, mock_request):
        self.assert_all_servers_fail(mock_request, self._service.head)

    @mock.patch('balast.service.requests.get', return_value=_MockResponse(500))
    def test_all_get_servers_500(self, mock_request):
        self.assert_all_servers_fail(mock_request, self._service.get)

    @mock.patch('balast.service.requests.post', return_value=_MockResponse(500))
    def test_all_post_servers_500(self, mock_request):
        self.assert_all_servers_fail(mock_request, self._service.post)

    @mock.patch('balast.service.requests.put', return_value=_MockResponse(500))
    def test_all_put_servers_500(self, mock_request):
        self.assert_all_servers_fail(mock_request, self._service.put)

    @mock.patch('balast.service.requests.patch', return_value=_MockResponse(500))
    def test_all_patch_servers_500(self, mock_request):
        self.assert_all_servers_fail(mock_request, self._service.patch)

    @mock.patch('balast.service.requests.delete', return_value=_MockResponse(500))
    def test_all_delete_servers_500(self, mock_request):
        self.assert_all_servers_fail(mock_request, self._service.delete)

    def assert_request_ok(self, mock_request, request_call):

        # make our request
        result = request_call('/relative/path')

        # should have tried twice - once for each server
        self.assertEqual(1, mock_request.call_count)
        self.assertEqual(result, mock_request.return_value)

        # verify the url we create
        call_arg = mock_request.call_args_list[0]
        url = UrlBuilder.from_url(call_arg[0][0])
        self.assertIn(url._hostname, _EXPECTED_SERVERS)
        self.assertEqual(url._scheme, 'https')
        self.assertIsNone(url._port)
        self.assertEqual(url._path, '/relative/path')

    def assert_all_servers_fail(self, mock_request, request_call):

        # will try each server, and mark each down on a 500 response
        # when no servers are left, a NoReachableServersException is thrown
        self.assertRaises(NoReachableServers, request_call, '/relative/path')

        # should have tried twice - once for each server
        self.assertEqual(2, mock_request.call_count)

        # add our hostnames to a set (to ensure the
        # same server isn't being used repeatedly)
        # all other info should be the same for each server request
        hostnames = set()
        for i, call_arg in enumerate(mock_request.call_args_list):
            url = UrlBuilder.from_url(call_arg[0][0])
            hostnames.add(url._hostname)
            self.assertEqual(url._scheme, 'https')
            self.assertIsNone(url._port)
            self.assertEqual(url._path, '/relative/path')

        # less than 2 means a server has been reused
        self.assertEqual(2, len(hostnames))
        for hostname in hostnames:
            self.assertIn(hostname, _EXPECTED_SERVERS)

    @staticmethod
    def _create_service():

        servers = StaticServerList(_EXPECTED_SERVERS)
        load_balancer = LoadBalancer(servers, ping=ping.DummyPing(), ping_on_start=False)
        load_balancer.ping()

        return Service(load_balancer, use_https=True, request_timeout=0.1)



import logging
import requests
from past.builtins import unicode
from balast.util import UrlBuilder
from balast.discovery import Server, ServerList


class ConsulRestRecordList(ServerList):

    _SERVICE_URL = '/v1/catalog/service/'

    def __init__(self, base_url, service, dc=None, near=None, tag=None):
        super(ConsulRestRecordList, self).__init__()
        self.base_url = base_url
        self.service = service
        self.dc = dc
        self.near = near
        self.tag = tag
        self._logger = logging.getLogger(self.__module__)

    def get_servers(self):

        try:
            url = UrlBuilder.from_url(self.base_url)
            url.path(self._SERVICE_URL)
            url.append_path(self.service)

            if self.dc is not None:
                url.add_query_param('dc', self.dc)

            if self.near is not None:
                url.add_query_param('near', self.near)

            if self.tag is not None:
                url.add_query_param('tag', self.tag)

            response = requests.get(unicode(url))
            json = response.json()

            for entry in json:
                s = Server(
                    entry['Address'],
                    entry['ServicePort'],
                    ttl=10
                )

                self._logger.debug("Created server from Consul REST API record: %s", s)

                yield s

        except:
            return



import unittest
import mock
import socket
import time
from balast import Server
from balast.discovery.static import StaticServerList
from balast.ping import (
    Ping,
    DummyPing,
    SocketPing,
    UrlPing,
    SerialPingStrategy,
    ThreadPoolPingStrategy,
    MultiprocessingPoolPingStrategy,
    GeventPingStrategy
)


class _MockPing(Ping):

    def __init__(self, is_alive=True, delay=0):
        super(_MockPing, self).__init__()
        self._is_alive = is_alive
        self._delay = delay

    def is_alive(self, server):
        time.sleep(self._delay)
        return self._is_alive


class DummyPingTest(unittest.TestCase):

    def test_is_dummy_alive(self):

        ping = DummyPing()
        server = Server('127.0.0.1', 80)

        # dummy ping assumes everything is alive
        self.assertTrue(ping.is_alive(server))


class SocketPingTest(unittest.TestCase):

    @mock.patch('balast.ping.socket.socket')
    def test_is_socket_alive_success(self, mock_socket):

        expected_ping_time = 3

        ping = SocketPing()
        ping.max_ping_time = expected_ping_time

        server = Server('127.0.0.1', 80)

        # ping our server via (mock) TCP socket
        self.assertTrue(ping.is_alive(server))
        mock_socket.assert_called_with(socket.AF_INET, socket.SOCK_STREAM)

        # verify the socket received the expected values
        s = mock_socket.return_value
        s.settimeout.assert_called_with(expected_ping_time)
        s.connect.assert_called_with(('127.0.0.1', 80))

    @mock.patch('balast.ping.socket.socket.connect')
    def test_is_socket_alive_fails(self, mock_connect):

        mock_connect.side_effect = Exception('mock exception')
        ping = SocketPing()

        server = Server('127.0.0.1', 80)

        # ping our server via (mock) TCP socket
        self.assertFalse(ping.is_alive(server))


class UrlPingTest(unittest.TestCase):

    @mock.patch('balast.ping.requests.get')
    def test_is_url_alive_success(self, mock_get_request):

        ping = UrlPing()
        server = Server('127.0.0.1', 80)
        expected_http_address = 'http://127.0.0.1:80'
        expected_https_address = 'https://127.0.0.1:80'

        # ping our server via (mock) HTTP request
        self.assertTrue(ping.is_alive(server))
        mock_get_request.assert_called_with(expected_http_address)

        # now do the same thing with https
        ping = UrlPing(is_secure=True)

        # ping our server via (mock) HTTP request
        self.assertTrue(ping.is_alive(server))
        mock_get_request.assert_called_with(expected_https_address)

    @mock.patch('balast.ping.requests.get')
    def test_is_url_alive_fails(self, mock_get_request):

        mock_get_request.side_effect = Exception('mock exception')
        ping = UrlPing()

        server = Server('127.0.0.1', 80)

        # ping our server via (mock) TCP socket
        self.assertFalse(ping.is_alive(server))


class SerialPingStrategyTest(unittest.TestCase):

    def test_serial_ping(self):

        servers = StaticServerList(['127.0.0.1', '127.0.0.2', '127.0.0.3'])
        ping = _MockPing(delay=0.4)
        strategy = SerialPingStrategy()

        results = strategy.ping(ping, servers)

        # verify the results
        self.assertEqual(3, len(results))

        for server in results:
            self.assertTrue(server.is_alive)


class ThreadPoolPingStrategyTest(unittest.TestCase):

    def test_thread_pool_ping(self):

        servers = StaticServerList([])
        for i in range(100):
            servers.add_server('127.0.0.%s' % i)

        ping = _MockPing(delay=0.4)
        strategy = ThreadPoolPingStrategy()

        results = strategy.ping(ping, servers)

        # verify the results
        self.assertEqual(100, len(results))

        for server in results:
            self.assertTrue(server.is_alive)


class MultiprocessingPoolPingStrategyTest(unittest.TestCase):

    def test_multiprocessing_pool_ping(self):

        servers = StaticServerList([])
        for i in range(10):
            servers.add_server('127.0.0.%s' % i)

        ping = _MockPing(delay=0.4)
        strategy = MultiprocessingPoolPingStrategy()

        results = strategy.ping(ping, servers)

        # verify the results
        self.assertEqual(10, len(results))

        for server in results:
            self.assertTrue(server.is_alive)


class GeventPingStrategyTest(unittest.TestCase):

    def test_gevent_pool_ping(self):

        servers = StaticServerList([])
        for i in range(10):
            servers.add_server('127.0.0.%s' % i)

        ping = _MockPing(delay=0.1)
        strategy = GeventPingStrategy()

        results = strategy.ping(ping, servers)

        # verify the results
        self.assertEqual(10, len(results))

        for server in results:
            self.assertTrue(server.is_alive)
